//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : CWAVFILE.CPP
//
// Description:
//		Implementation of class CWAVFile. This class reads and writes a SOUNDBLASTER
//		WAV-file format.
//
//////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <math.h>

#include "defines.h"
#include "ownerr.h"
#include "CWAVFile.h"

// constants
#define FALSE				0
#define TRUE				1
#define PI					3.14159265

// some usedfull macros
#define MIN(a,b)	(a < b ? a : b)
#define MAX(a,b)	(a > b ? a : b)


//////////////////////////////////////////////////////////////////////////////
// PUBLIC METHODS
//////////////////////////////////////////////////////////////////////////////

CWAVFile::CWAVFile()
{
	memset(&WAVHeader,0x00,sizeof(WAVHeader));
	fpWAV=NULL;
	SetToneFunction(DEF_TONEFUNCTION);
}

CWAVFile::CWAVFile(char *filename)
{

	// filename is assumed as input file name
	SetFileName(filename);
	if (OpenWAV() != ERR_WAV_NOERROR)
		ReadWAV();
	else
		fpWAV=NULL;
	SetToneFunction(DEF_TONEFUNCTION);
}

CWAVFile::~CWAVFile()
{
	CloseWAV();
}

WAV_ERR_STATUS CWAVFile::OpenWAV(void)
{
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;

	if ((fpWAV=fopen(FileName,"rb+"))==NULL)
	{
		status=ERR_WAV_OPEN;
		PrintError(status);
	}				
	return(status);
}

WAV_ERR_STATUS CWAVFile::CloseWAV(void)
{
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;

	if (fpWAV!=NULL)
	{
		if ((status=WriteHeader())!=ERR_WAV_NOERROR)
			PrintError(ERR_WAV_WRITEHEADER);
		if (fclose(fpWAV)!=NULL)
		{
			status=ERR_WAV_CLOSE;
			PrintError(status);
		}
	}
	return(status);
}

WAV_ERR_STATUS CWAVFile::ReadWAV(void)
{
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;

	if ((status=ReadHeader())!=ERR_WAV_NOERROR)	
		return(status);
	return(status);
}

void CWAVFile::SetValue(char channel,DWORD index, WORD value)
{
	FILE *helpFP;
	char wordsize=WAVHeader.Bits_P_Sample;

	FSeekToSample(channel,index);

	switch(wordsize)
	{
		case 8:
			fputc((char) (value & 0x00FF),fpWAV);
			break;
		case 12:
		case 16:
			fwrite(&value,sizeof(WORD),1,fpWAV);
			break;
		default:
			PrintError(ERR_WAV_ILLEGAL);
			return;
	}
}

WORD CWAVFile::GetValue(char channel,DWORD index)
{
	WORD value;
	char wordsize=WAVHeader.Bits_P_Sample;

	FSeekToSample(channel,index);

	switch(wordsize)
	{
	   case 8:
			value=(WORD) fgetc(fpWAV);
			value&=0x00FF;
			break;
		case 12:
	   case 16:
			fread(&value,sizeof(WORD),1,fpWAV);
			break;
		default:
			PrintError(ERR_WAV_ILLEGAL);
			value=0x00;
	}
	return(value);
}

void CWAVFile::SetToneFunction(char *funcsrc)
{
	ToneFunction.NewString(funcsrc);
}

void CWAVFile::SetToneFunctionFromFile(char *filename)
{
	ToneFunction.NewFile(filename);
}

WORD CWAVFile::OptimizeToneFunction(void)
{
	return(ToneFunction.Optimize());
}

void CWAVFile::SetToneFunctionParameter(char *parameter,VALUETYPE value)
{
	if (strlen(parameter)!=1)
		warning("In this version only single character variablenames are allowed.\nTaking the first letter only.");
	ToneFunction.SetParameter(parameter[0],value);
}

void CWAVFile::SetMode(WORD mode)
{
	if ((mode!=1) && (mode!=2))
	{
		PrintError(ERR_WAV_ILLEGAL);
		WAVHeader.Mode=1;
		return;
	}
	WAVHeader.Mode=mode;
}

void CWAVFile::SetSampleRate(DWORD samplerate)
{
	if ((samplerate!=SAMPLERATE1) &&
       (samplerate!=SAMPLERATE2) &&
       (samplerate!=SAMPLERATE3))
	{
		PrintError(ERR_WAV_ILLEGAL);
		WAVHeader.SampleRate=SAMPLERATE1;
		return;
	}
	WAVHeader.SampleRate=samplerate;
}

WAV_ERR_STATUS CWAVFile::NewFile(char  *filename,
                                 WORD  mode,
                                 DWORD samplerate,
                                 DWORD bytes_p_sec,
                                 WORD  bytes_p_sample,
                                 WORD  bits_p_sample,
                                 VALUETYPE  noofseconds)
{
	// filename is assumed as output file name
	SetFileName(filename);
	// constant header data
	strcpy(WAVHeader.TextRIFF,"RIFF");
	strcpy(WAVHeader.TextWAVE,"WAVE");
	strcpy(WAVHeader.TextFMT ,"fmt ");
	strcpy(WAVHeader.TextDATA,"data");
	WAVHeader.Sub_Chunk_Length=0x10;
	WAVHeader.Format=0x01;

	SetMode(mode);
	SetSampleRate(samplerate);
	SetBytesPerSecond(bytes_p_sec);

	WAVHeader.Bytes_P_Sample=bytes_p_sample;
	WAVHeader.Bits_P_Sample =bits_p_sample;

	return(NewMem(noofseconds));
}

WAV_ERR_STATUS CWAVFile::NewMem(VALUETYPE noofseconds)
{
	DWORD size;
	DWORD cnt;
	DWORD realsize;
	WORD  zerovalue;
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;

	size=(DWORD) (noofseconds*WAVHeader.Bytes_P_Sec*WAVHeader.Mode);
	size+=size % 2;

	switch(WAVHeader.Bits_P_Sample)
	{
		case 8:
			realsize=size;
			WAVHeader.Bytes_P_Sample=1;
			zerovalue=0x7F;
			break;
		case 12:
			realsize=2*size;
			WAVHeader.Bytes_P_Sample=2;
			zerovalue=0x07FF;
			break;
		case 16:
			realsize=2*size;
			WAVHeader.Bytes_P_Sample=2;
			zerovalue=0x7FFF;
			break;
		default:
			PrintError(ERR_WAV_ILLEGAL);
			return(ERR_WAV_ILLEGAL);
	}
	WAVHeader.DataChunkLength=realsize;

	status=CreateWAVFile(realsize+sizeof(WAVHeader),zerovalue);
	return(status);
}


void CWAVFile::GenTone(char channel,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE start,VALUETYPE length,VALUETYPE compression)
{
	DWORD startindex;
	DWORD endindex;
	VALUETYPE factor;
	DWORD t;
	DWORD value;
	DWORD zerooffset;
	WORD w_amplitude;
	VALUETYPE f_value;

	startindex=MIN((DWORD) (start*WAVHeader.Bytes_P_Sec),WAVHeader.DataChunkLength);
	endindex  =MIN(startindex+(DWORD) (length*WAVHeader.Bytes_P_Sec),WAVHeader.DataChunkLength);
	factor    =(WAVHeader.SampleRate/frequency);

	switch(WAVHeader.Bits_P_Sample)
	{
		case 8:
			zerooffset=127;
			break;
		case 12:
		case 16:
			zerooffset=0;
		break;
	}
	
	w_amplitude=(VALUETYPE) ((1 << WAVHeader.Bits_P_Sample-1)-1)*amplitude;
	ToneFunction.SetParameter(V_AMPLITUDE,w_amplitude);
	ToneFunction.SetParameter(V_SINWT_T,(VALUETYPE) factor);

	for (t=0; t<(endindex-startindex); t++)
	{
		ToneFunction.SetParameter(V_SAMPLEINDEX,(VALUETYPE) t);
		f_value=compression*(VALUETYPE) ToneFunction(t);
		value=(DWORD) f_value+zerooffset;
		SetValue(channel,startindex+t,value);
	}
}

VALUETYPE CWAVFile::CalcToneCompression(char channel,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE start,VALUETYPE length,VALUETYPE compression)
{
	DWORD startindex;
	DWORD endindex;
	DWORD factor;
	DWORD t;
	WORD  w_amplitude;
	VALUETYPE f_value;
	VALUETYPE new_compression;
	VALUETYPE tmp_new_compression;
	VALUETYPE max_value;

	channel=channel;  // TO AVOID COMPILER WARNING

	startindex=MIN((DWORD) (start*WAVHeader.Bytes_P_Sec),WAVHeader.DataChunkLength);
	endindex  =MIN(startindex+(DWORD) (length*WAVHeader.Bytes_P_Sec),WAVHeader.DataChunkLength);
	factor    =(DWORD) (WAVHeader.SampleRate/frequency);

	max_value=(VALUETYPE) ((1 << WAVHeader.Bits_P_Sample-1)-1);
	w_amplitude=max_value*amplitude;

	ToneFunction.SetParameter(V_AMPLITUDE,w_amplitude);
	ToneFunction.SetParameter(V_SINWT_T,factor);

	new_compression=compression;
	for (t=0; t<(endindex-startindex); t++)
	{
		f_value=compression*(VALUETYPE) ToneFunction(t);
		if (abs(f_value) > max_value)
		{
			tmp_new_compression=max_value/abs(f_value);
			if (tmp_new_compression < new_compression)
				new_compression=tmp_new_compression;
		}
	}
	return(new_compression);
}

void CWAVFile::PrintError(WAV_ERR_STATUS errorid)
{
	char *errormsg=alloc_msgmem();

	strcpy(errormsg,"WAV: ");
	switch(errorid)
	{
		case ERR_WAV_NOERROR:
			strcat(errormsg,"No errors.");
			break;
		case ERR_WAV_OPEN:
			strcat(errormsg,"Cannot open WAV-file.");
			break;
		case ERR_WAV_CLOSE:
			strcat(errormsg,"Cannot close WAV-file.");
			break;
		case ERR_WAV_HEADER:
			strcat(errormsg,"Error in HEADER.");
			break;
		case ERR_WAV_FMTCHUNK:
			strcat(errormsg,"Error in FORMAT-CHUNK.");
			break;
		case ERR_WAV_DATACHUNK:
			strcat(errormsg,"Error in DATA-CHUNK.");
			break;
		case ERR_WAV_OUTOFMEM:
			strcat(errormsg,"Out of memory.");
			break;
		case ERR_WAV_ILLEGAL:
			strcat(errormsg,"Illegal parameter - using default.");
			break;
		case ERR_WAV_AMPOVERFLOW:
			strcat(errormsg,"Amplitude to big.");
			break;
		case ERR_WAV_WRITEHEADER:
			strcat(errormsg,"Cannot write WAV-header to WAV-file.");
			break;
		case ERR_WAV_EXPORTOPEN:
			strcat(errormsg,"Cannot write WAV-header to WAV-file.");
			break;
	}
	error(errormsg);
	dealloc_msgmem(errormsg);
}


WAV_ERR_STATUS CWAVFile::ExportAsInteger(char channel,char *filename,char normalized)
{
	DWORD cnt;
	FILE *fpTEXT;
	WORD offset=0;
	char printfmt[16]="%u\n";

	if ((fpTEXT=fopen(filename,"w"))==NULL)
		return(ERR_WAV_EXPORTOPEN);
	if (normalized==FILE_NORMALIZED)
	{
		strcpy(printfmt,"%d\n");
		offset=(WORD) (pow(2,WAVHeader.Bits_P_Sample-1)) - 1;
	}
	for (cnt=0; cnt<WAVHeader.DataChunkLength; cnt++)
		fprintf(fpTEXT,printfmt,GetValue(channel,cnt)-offset);
	fclose(fpTEXT);

	return(ERR_WAV_NOERROR);
}

//////////////////////////////////////////////////////////////////////////////
// PROTECTED METHODS
//////////////////////////////////////////////////////////////////////////////

WAV_ERR_STATUS CWAVFile::CreateWAVFile(DWORD size,char zerovalue)
{
	DWORD i;
	WAV_ERR_STATUS status = ERR_WAV_NOERROR;

	// if an existing file is open.
	// this file is erased.
	CloseWAV();
	if ((fpWAV=fopen(FileName,"wb+"))==NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"Cannot open file %s\n",FileName);
		fatal(msg);
		dealloc_msgmem(msg);
		status=ERR_WAV_OPEN;
	}
	for (i=0; i<size; i++)
		fputc(zerovalue,fpWAV);

	// this newly created file remains open !!
	return(status);
}

DWORD CWAVFile::CalcSampleIndex(char channel,DWORD index)
{
	DWORD newindex;

	newindex=index*WAVHeader.Bytes_P_Sample*WAVHeader.Mode;
	switch(channel)
	{
		case CHANNEL_MONO:
			break;
	   case CHANNEL_LEFT:
			break;
		case CHANNEL_RIGHT:
			newindex+=WAVHeader.Bytes_P_Sample;
			break;
	}
	return(newindex);
}

// NOT YET CORRECT !!!!!!!!!!!!!!!!!!!!!!!!!
void CWAVFile::FSeekToSample(char channel,DWORD index)
{
	fseek(fpWAV,sizeof(WAVHeader)+CalcSampleIndex(channel,index),SEEK_SET);
}

WAV_ERR_STATUS CWAVFile::ReadHeader(void)
{
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;
	char textRIFF[5];
	char textWAVE[5];
	char textFMT[5];
	char textDATA[5];
	int i;

	fread(&WAVHeader,sizeof(WAVHeader),1,fpWAV);

	for (i=0;i<4;i++)
	{
		textRIFF[i]=WAVHeader.TextRIFF[i];
		textWAVE[i]=WAVHeader.TextWAVE[i];
		textFMT[i] =WAVHeader.TextFMT[i];
		textDATA[i]=WAVHeader.TextDATA[i];
	}
	textRIFF[4]=textWAVE[4]=textFMT[4]=textDATA[4]='\0';

	#ifdef DEBUG
		printf("READING FILE:\n");
		PrintDebugInfo(1);
	#endif

	if ((strcmp(textRIFF,"RIFF") |
		  strcmp(textWAVE,"WAVE") |
		  strcmp(textFMT ,"fmt ") |
		  strcmp(textDATA,"data"))!=0)
		status=ERR_WAV_HEADER;

	return(status);
}

WAV_ERR_STATUS CWAVFile::WriteHeader(void)
{
	WAV_ERR_STATUS status=ERR_WAV_NOERROR;

	WAVHeader.Bytes_P_Sample*=WAVHeader.Mode;
	//WAVHeader.DataChunkLength*=WAVHeader.Mode;
	
	WAVHeader.Length=WAVHeader.DataChunkLength+44;

	// when debugging is enabled, all parts of the header is displayed
	#ifdef DEBUG
		printf("WRITING FILE:\n");
		PrintDebugInfo(2);
	#endif

	fseek(fpWAV,0,SEEK_SET);
	fwrite(&WAVHeader,sizeof(WAVHeader),1,fpWAV);

	return(status);
}

// when debugging is enabled, all parts of the header is displayed
#ifdef DEBUG
	void CWAVFile::PrintDebugInfo(char readwrite)
	{
		char textRIFF[5];
		char textWAVE[5];
		char textFMT[5];
		char textDATA[5];
		int i;

		for (i=0;i<4;i++)
		{
			textRIFF[i]=WAVHeader.TextRIFF[i];
			textWAVE[i]=WAVHeader.TextWAVE[i];
			textFMT[i] =WAVHeader.TextFMT[i];
			textDATA[i]=WAVHeader.TextDATA[i];
		}
		textRIFF[4]=textWAVE[4]=textFMT[4]=textDATA[4]='\0';

		printf("-----------------------------------------------------------------------------\n");
		if (readwrite==1)
			printf("INPUT-FILE: %s\n",InFileName);
		else
			printf("OUTPUT-FILE: %s\n",OutFileName);
		printf("HEADER:\n");
		printf("[string]         = %s\n",textRIFF);
		printf("filelength       = %ld , 0x%lx\n",WAVHeader.Length,WAVHeader.Length);
		printf("[string]         = %s\n",textWAVE);
		printf("FMTCHUNK:\n");
		printf("[string]         = %s\n",textFMT);
		printf("chunk length     = %ld , 0x%lx\n",WAVHeader.Sub_Chunk_Length,WAVHeader.Sub_Chunk_Length);
		printf("format(PCM)      = %d\n",WAVHeader.Format);
		printf("mono(1)/stereo(2)= %d\n",WAVHeader.Mode);
		printf("samplerate       = %ld , 0x%lx\n",WAVHeader.SampleRate,WAVHeader.SampleRate);
		printf("bytes/second     = %ld , 0x%lx\n",WAVHeader.Bytes_P_Sec,WAVHeader.Bytes_P_Sec);
		printf("bytes/sample     = %d , 0x%x\n",WAVHeader.Bytes_P_Sample,WAVHeader.Bytes_P_Sample);
		printf("bits/sample      = %d , 0x%x\n",WAVHeader.Bits_P_Sample,WAVHeader.Bits_P_Sample);
		printf("DATACHUNK:\n");
		printf("[string]         = %s\n",textDATA);
		printf("chunk length     = %ld , 0x%0lx\n",WAVHeader.DataChunkLength,WAVHeader.DataChunkLength);
		printf("-----------------------------------------------------------------------------\n");
}
#endif
