//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : CWAVFILE.H
//
// Description:
//		Definition of class CWAVFile. This class reads and writes a SOUNDBLASTER
//		WAV-file format.
//
//////////////////////////////////////////////////////////////////////////////

// for debugging uncomment the following line
// #define DEBUG

#ifndef __CWAVFILE
#define __CWAVFILE

#include <string.h>
#include <stdio.h>

#include <ccpptyp.h>

#include "defines.h"
#include "Function.h"

// data format definitions
#define WAV_ERR_STATUS				unsigned char

// SAMPLERATE definitions	
#define SAMPLERATE1					11025
#define SAMPLERATE2					22050
#define SAMPLERATE3					44100

// mode definitions
#define MODE_MONO						0x01
#define MODE_STEREO					0x02

// import, export file operation definition
#define FILE_NOT_NORMALIZED 		0x01
#define FILE_NORMALIZED		 		0x02

// channel definitions
#define CHANNEL_MONO					0x01
#define CHANNEL_LEFT					0x02
#define CHANNEL_RIGHT				0x03

// tone generation modes - only used in 'GenTone'
#define GENTONE_NEW					0x01

// default tone function
#define DEF_TONEFUNCTION "main { return a*(sin(6.283185307/f*x)); }"

// parameter defaults
#define DEF_MODE						1						// mono
#define DEF_SAMPLERATE				0x2B11				// 11025 Hz
#define DEF_BYTES_P_SEC				0x2B11				// 1 second sound
#define DEF_BYTES_P_SAMPLE 		1						// 8 bit
#define DEF_BITS_P_SAMPLE			8						// 8 bit
#define DEF_NOOFSECONDS				1						// 1 second

// other parameter defaults
#define DEF_COMPRESSION				1.0

#define MAXPATHLEN					128

// ERROR definitions
#define ERR_WAV_NOERROR				0x80
#define ERR_WAV_OPEN					0x81
#define ERR_WAV_CLOSE				0x82
#define ERR_WAV_HEADER				0x83
#define ERR_WAV_FMTCHUNK			0x84
#define ERR_WAV_DATACHUNK			0x85
#define ERR_WAV_OUTOFMEM			0x86
#define ERR_WAV_ILLEGAL				0x87
#define ERR_WAV_AMPOVERFLOW		0x88
#define ERR_WAV_WRITEHEADER		0x89
#define ERR_WAV_EXPORTOPEN			0x90

typedef struct _WAVHEADER
{
	// header
	char	TextRIFF[4];			// 'RIFF'
	DWORD Length;					// filesize
	char	TextWAVE[4];			// 'WAVE'

	// FMT-Chunk
	char	TextFMT[4];				// 'fmt '
	DWORD	Sub_Chunk_Length;		// length of sub_chunk; allways = 16 bytes
	WORD	Format;					// at this time = 1 for PCM-code
	WORD	Mode;						// mono = 1; stereo = 2
	DWORD	SampleRate;				// sample-frequency
	DWORD	Bytes_P_Sec;			// bytes/second
	WORD	Bytes_P_Sample;		// bytes/sample; 8 bit = 1, 16 bit = 2
	WORD	Bits_P_Sample;			// bit/sample = [8,12,16]

	// DATA-Chunk
	char	TextDATA[4];			// 'data'
	DWORD DataChunkLength;		//	length of data-block
};

class CWAVFile
{
	// WAV-file related data
	char FileName[MAXPATHLEN];
	FILE *fpWAV;

	// function for tone-generation
	Function ToneFunction;

	// header data
	_WAVHEADER	WAVHeader;

public:
	// constructors
	CWAVFile();
	CWAVFile(char *filename);
	// destructor
	~CWAVFile();

	// FILE NAME OPERATIONS
	// set filename for reading
	void SetFileName(char *filename) 	{ strcpy(FileName,filename);  }

	// FILE OPERATIONS
	// open WAV-File
	WAV_ERR_STATUS OpenWAV(void);						
	// close WAV-File
	WAV_ERR_STATUS CloseWAV(void);						

	// READ OPERATIONS
	// read WAV-parameters into header-struct
	WAV_ERR_STATUS ReadWAV(void);						

	// VALUE OPERATIONS
	// get a single value in DATA_Chunk
	WORD GetValue(char channel,DWORD index);
	// modifies a single value in DATA-Chunk
	void SetValue(char channel,DWORD index, WORD value);

	// FUNCTION OPERATIONS
	void SetToneFunction(char *funcsrc);
	void SetToneFunctionFromFile(char *filename);
	void SetToneFunctionParameter(char *parameter,VALUETYPE);
	WORD OptimizeToneFunction(void);

	// PARAMETER OPERATIONS
	// set format (mono/stereo)
	void SetMode(WORD mode);
	// get format (mono/stereo)
	WORD GetMode(void)								{ return(WAVHeader.Mode);		  	}
	// set samplerate
	void SetSampleRate(DWORD samplerate);
	// get samplerate
	DWORD GetSampleRate(void)						{ return(WAVHeader.SampleRate);  }
	// set  no of bytes/second
	void SetBytesPerSecond(DWORD bytes_p_sec)	{ WAVHeader.Bytes_P_Sec=bytes_p_sec; }
	// get  no of bytes/second
	DWORD GetBytesPerSecond(void)					{ return(WAVHeader.Bytes_P_Sec); }
	// get size of DATA-Chunk
	DWORD GetDataChunkLength(void) { return WAVHeader.DataChunkLength; }


	// SOUND OPERATIONS
	// generate new sound info (incl. headerinfo)
	// parameters: possible values
	//	mode           ... 1=mono, 2=stereo;
	//	samplerate     ... 11025=0x2B11, 22050=0x5622, 44100=0xAC44 [Hz]
	// bytes_p_sec    ...
	// bytes_p_sample	... 1 = 8 bit, 2 = 16 bit
	// bits_p_sample  ... 8,12,16 bit
	// noofseconds		... n [sec]

	WAV_ERR_STATUS NewFile(char  *filename,
								  WORD  mode          =DEF_MODE,
								  DWORD samplerate    =DEF_SAMPLERATE,
								  DWORD bytes_p_sec   =DEF_BYTES_P_SEC,
								  WORD  bytes_p_sample=DEF_BYTES_P_SAMPLE,
								  WORD  bits_p_sample =DEF_BITS_P_SAMPLE,
								  VALUETYPE noofseconds   =DEF_NOOFSECONDS);
	// allocates or reallocates and initializes DataChunk for n number of seconds
	WAV_ERR_STATUS NewMem(VALUETYPE noofseconds);

	// generate a tone with
	//		frequency [Hz],
	//		amplitude,
	//		start [sec],
	//		length [sec],
	//    compression
	void GenTone(char channel,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE start,VALUETYPE length,VALUETYPE compression);

	// calculate new compression ratio for GenTone so that the signal does not
	// exceed the upper and lower limits
	VALUETYPE CalcToneCompression(char channel,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE start,VALUETYPE length,VALUETYPE compression);

	// PRINTOUT OPERATIONS
	void PrintError(WAV_ERR_STATUS errorid);

	// ASCII OUTPUT OPERATIONS
	// outputs whole DataChunk to a text file as integer values
	WAV_ERR_STATUS ExportAsInteger(char channel,char *filename,char normalized);

protected:
   ///////////////////////////////////////////////////////////////////////////
	// FILE CREATION
   ///////////////////////////////////////////////////////////////////////////
	// create a new WAV-file
	WAV_ERR_STATUS CreateWAVFile(DWORD size,char zerovalue);

   ///////////////////////////////////////////////////////////////////////////
	// INDEX CALCULATION AND SEEKING
   ///////////////////////////////////////////////////////////////////////////
	// calculate starting byte index of a sample
	DWORD CalcSampleIndex(char channel,DWORD index);
	// seek to specific position in DATAChunk
	void FSeekToSample(char channel,DWORD index);

   ///////////////////////////////////////////////////////////////////////////
	// READ OPERATIONS - used by ReadAll;
   ///////////////////////////////////////////////////////////////////////////
	WAV_ERR_STATUS ReadHeader(void);					// reads and evaluates header
   ///////////////////////////////////////////////////////////////////////////

   ///////////////////////////////////////////////////////////////////////////
	// WRITE OPERATIONS - used by WriteAll;
   ///////////////////////////////////////////////////////////////////////////
	WAV_ERR_STATUS WriteHeader(void);				// writes the header
   ///////////////////////////////////////////////////////////////////////////

   ///////////////////////////////////////////////////////////////////////////
	// DEBUGGING METHODS
   ///////////////////////////////////////////////////////////////////////////
	// print debug info
	// parameter readwrite : 1=read  operation
	// 							 2=write operation
	void PrintDebugInfo(char readwrite);
};

#endif
