%{
//////////////////////////////////////////////////////////////////////////////
//
// DIPLOMARBEIT : WAVE-GL 	Inst. f. techn. Mathematik / TU-WIEN
//
// Author : Michael Mueller-Wernhart
//
//////////////////////////////////////////////////////////////////////////////
//
// MODULE : WAVPARS.Y
//
// Description:
//		Parser of WAV-generation language.
//
//	Version 1.0
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// special comments
//
//	'ps->' in function-calls warning,error and fatal references to the scanner.
//        This pointer is automatically created by YACC.
//////////////////////////////////////////////////////////////////////////////
#include <ctype.h>

#include "defines.h"
#include "ownerr.h"
#include "CWAVFile.h"
#include "WavScan.h"
#include "WavPars.h"
#include "SimplVar.h"
#include "HashTab.h"

// verbose modes
#define SWITCH_OFF	0x00
#define SWITCH_ON		0x01

// some error codes

// variables
CWAVFile WAVFile;
HashTable WAV_LongVariables;

char function_defined=FALSE;

// switches
char verbose=SWITCH_OFF;
char autoscale=SWITCH_OFF;

// prototypes
void generate_tone_proc(char channelid,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE startat, VALUETYPE duration);
void WAV_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno);

%}

%union {
   VALUETYPE fvalue;
   unsigned long ivalue;
	char *svalue;
};

/* tokens without attributes */
%token AUTOSCALE
%token BIT BUFFERSIZE
%token CHANNEL CREATE
%token WAV_DECLARE DEFAULTS DEFFUNC
%token S_END EXPORT
%token FROM
%token GENERATE
%token HZ
%token IMPORT IS
%token KHZ
%token MSEC
%token LEFTCH
%token MESSAGE MONO
%token NEW NORMALIZED
%token ON OFF
%token RIGHTCH
%token SAMPLERATE SEC S_FILE STEREO
%token TO TONE TRANSPORT
%token USING
%token VERBOSE
%token WITH

/* tokens with attribute */
%token <fvalue> FLOATVALUE
%token <ivalue> INTVALUE
%token <svalue> WAV_LONGVARIABLE
%token <svalue> STRING
%token <svalue> WAV_VARIABLE

%type	<fvalue> amplitude
%type <ivalue> bits
%type <ivalue> channelid
%type <fvalue> duration
%type <fvalue> duration_multiplier
%type <fvalue> expression
%type <fvalue> frequency
%type <fvalue> frequency_multiplier
%type <fvalue> int_or_float
%type <ivalue> mono_stereo
%type <ivalue> normalized
%type <ivalue> on_off
%type <ivalue> samplerate
%type <fvalue> startat

%left '+' '-'
%left '*' '/'

%start program

%%


program:
	  opt_verbose create_new_file lines S_END
	  {
	 		if (verbose==SWITCH_ON)
				printf("\nTerminated successfully.\n");
	  }
	| opt_verbose create_new_file S_END
	  {
	 		warning("WAVPARS: new file created, but there was nothing to do !\n",ps->wavlineno);
	  }
	| opt_verbose S_END
	  {
	 		warning("WAVPARS:there was absolutly nothing to do !\n",ps->wavlineno);
	  }
	;

opt_verbose:
	  verbose ';'
	| /* empty */
	;

lines:
	  line delimiter
	| lines line delimiter
	;

line:
	  autoscale
	| declare
	| definefunc
	| export_import
	| generate_tone
	| longvariable_assignment
	| message
	| parameters
	| variable_assignment
	| verbose
	;

create_new_file:
	  NEW S_FILE STRING USING DEFAULTS ';'
	  {
	 		if (verbose==SWITCH_ON)
			{
				printf("Generating new file %s using defaults:\n",$3);
				printf("   mono(1)/stereo(2) : %7d\n",DEF_MODE);
				printf("   samplerate        : %7u Hz\n",DEF_SAMPLERATE);
				printf("   bits/sample       : %7d bit\n",DEF_BITS_P_SAMPLE);
				printf("   duration          : %7.3f sec\n",DEF_NOOFSECONDS);
			}
	 		WAVFile.NewFile($3);
			delete $3;	// remove string
	  }
	| NEW S_FILE STRING mono_stereo ',' samplerate ',' bits ',' duration ';'
	  {
	      /* bytes_p_sample=(mono_stereo * bits)/8 */
	 		unsigned int bytes_p_sample=bytes_p_sample=($4*$8)/8;		

	 		if (verbose==SWITCH_ON)
			{
				printf("Generating new file %s :\n",$3);
				printf("   mono(1)/stereo(2) : %7d\n",$4);
				printf("   samplerate        : %7u Hz\n",$6);
				printf("   bits/sample       : %7d bit\n",$8);
				printf("   duration          : %7.3f sec\n",$10);
			}
	 		WAVFile.NewFile($3,$4,$6,$6,bytes_p_sample,$8,$10);
			delete $3;	// remove string
	  }
	;

autoscale:
	  AUTOSCALE on_off
 	  {
			autoscale=$2;
	  }
	;

declare:
	  WAV_DECLARE declare_list
	;

declare_list:
	  declare_item
	| declare_list ',' declare_item
	;

declare_item:
     WAV_LONGVARIABLE
	  {
			WAV_declare_longvariable($1,0.0,ps->wavlineno);
			delete $1;
	  }
   | WAV_LONGVARIABLE '=' expression
	  {
			WAV_declare_longvariable($1,$3,ps->wavlineno);
			delete $1;
	  }
	;

definefunc:
	  DEFFUNC STRING
 	  {
	 		FKT_errorcnt=0;					// reset error count
		  	function_defined=TRUE;
			WORD nooptimizations;

			if (verbose==SWITCH_ON)
				printf("Compiling function...\n");
	 		WAVFile.SetToneFunction($2);

			if (verbose==SWITCH_ON)
				printf("Optimizing function...");
			// nooptimizations=WAVFile.OptimizeToneFunction();
			if (verbose==SWITCH_ON)
				//printf(" --> %d optimizations found.\n",nooptimizations);
				printf(" --> not available in this version!\n");
			delete $2; // remove string
	  }
	| DEFFUNC FROM S_FILE STRING
	  {
	 		FKT_errorcnt=0;					// reset error count
		  	function_defined=TRUE;
			WORD nooptimizations;

			if (verbose==SWITCH_ON)
				printf("Compiling function...\n");
	 		WAVFile.SetToneFunctionFromFile($4);

			if (verbose==SWITCH_ON)
				printf("Optimizing function...");
			// nooptimizations=WAVFile.OptimizeToneFunction();
			if (verbose==SWITCH_ON)
				//printf(" --> %d optimizations found.\n",nooptimizations);
				printf(" --> not available in this version!\n");
			delete $4; // remove string
	  }
	;

export_import:
	  EXPORT channelid CHANNEL TO S_FILE STRING normalized
	  {
	 		if (verbose==SWITCH_ON)
			{
				printf("Exporting data to file %s ",$6);
				if ($7==FILE_NORMALIZED)
					printf("normalized ");
				printf("...\n");
			}

	 		WAVFile.ExportAsInteger($2,$6,$7);
			delete $6; // remove string
	  }
	| IMPORT channelid CHANNEL FROM S_FILE STRING
	  {
	 		warning("WAVPARS: IMPORT not yet implemented !\n",ps->wavlineno);
			delete $6; // remove string
	  }
	;

generate_tone:
	  GENERATE TONE channelid ',' frequency ',' amplitude ',' startat ',' duration
	  {
	 		if (FKT_errorcnt>0)
				error("WAVPARS: tone cannot be generated, because of too many errors in function!");
			else
			{
				if (function_defined==FALSE)
					warning("WAVPARS: no tone-function defined !! Using default function a*sin(wt)",ps->wavlineno);
				generate_tone_proc($3,$5,$7,$9,$11);
			}
	  }
	;

longvariable_assignment:
	  WAV_LONGVARIABLE '=' expression
	  {
	 		SimpleVariable *help;

			if ((help=(SimpleVariable *) WAV_LongVariables.Search($1,VAR_SIMPLE))==NULL)
			{
				char *msg=alloc_msgmem();
				sprintf(msg,"WAVPARS: variable %s not declared!!",$1);
				fatal(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				help->SetValue($3);
			delete $1;
     }
	;

normalized:
	  NORMALIZED
	  {
	 		$$=FILE_NORMALIZED;
	  }
	| /* empty */
	  {
	 		$$=FILE_NOT_NORMALIZED;
	  }
	;

message:
	  MESSAGE STRING
 	  {
			printf("%s\n",$2);
			delete $2; // remove string
	  }
	;

parameters:
	  BUFFERSIZE '(' expression SEC ')'
	  {
			WAVFile.NewMem((float) $3);
	  }
	;

variable_assignment:
	  WAV_VARIABLE '=' expression
	  {
	      if (function_defined==FALSE)
			{
				char *msg=alloc_msgmem();

				sprintf(msg,"WAVPARS: Assignment of variable '%c' in default function a*sin(wt).",$1[0]);
				warning(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			if ((toupper($1[0])==toupper(V_SINWT_T))   	||
				 (toupper($1[0])==toupper(V_AMPLITUDE)) 	||
				 (toupper($1[0])==toupper(V_PARAMETER)) 	||
				 (toupper($1[0])==toupper(V_SAMPLEINDEX)) )
			{
				char *msg=alloc_msgmem();

				sprintf(msg,"WAVPARS: Ignoring assignment of reserved variable '%c' (See WAVCFG.CFG).",$1[0]);
				warning(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				WAVFile.SetToneFunctionParameter($1,$3);
	  }
	;

verbose:
	  VERBOSE on_off
	  {
	 		verbose=$2;
	  }
	;

/* utility rules */

expression:
	  int_or_float
	  {
	 		$$=$1;
	  }
	| expression '+' expression
	  {
	 		$$=$1 + $3;
	  }
	| expression '-' expression
	  {
	 		$$=$1 - $3;
	  }
	| expression '*' expression
	  {
	 		$$=$1 * $3;
	  }
	| expression '/' expression
	  {
	 		if ($3==0.0)
				fatal("WAVPARS: division by zero !!",ps->wavlineno);
	 		$$=$1 / $3;
	  }
   ;

int_or_float:
	  INTVALUE
	  {
	 		$$=(VALUETYPE) $1;
	  }
	| FLOATVALUE
	  {
	 		$$=$1;
	  }
	| WAV_LONGVARIABLE
	  {
	 		SimpleVariable *help;

			if ((help=(SimpleVariable *) WAV_LongVariables.Search($1,VAR_SIMPLE))==NULL)
			{
				char *msg=alloc_msgmem();
				sprintf(msg,"WAVPARS: variable %s not declared!!",$1);
				fatal(msg,ps->wavlineno);
				dealloc_msgmem(msg);
			}
			else
				$$=help->GetValue();
	  }
	;

mono_stereo:
	  MONO
 	  {
			$$=MODE_MONO;
	  }
	| STEREO
 	  {
			$$=MODE_STEREO;
	  }
	;

samplerate:
	  frequency
	  {
	 		$$=(unsigned int) $1;
	  }
	;

frequency:
	  expression frequency_multiplier
	  {
	 		$$=$1*$2;
	  }
	;

frequency_multiplier:
	  HZ
	  {
	 		$$=1.0;
	  }
	| KHZ
	  {
 			$$=1000.0;
	  }
	;

amplitude:
	  expression
	  {
	 		if (($1 < 0.0) || ($1 > 1.0))
			{
				error("WAVPARS: Amplitude must be between 0 and 1\nTone will NOT be generated !!",ps->wavlineno);
				$1=0.0;
			}
			else
				$$=$1;
	  }
	| expression '%'
	  {
	 		if ($1 > 100.0)
			{
				error("WAVPARS: Amplitude in % must be between 0 and 100\nTone will NOT be generated !!",ps->wavlineno);
				$1=0.0;
			}
			else
				$$=$1/100.0;
	  }
	;

bits:
	  INTVALUE BIT
	  {
	 		switch ($1)
			{
				case 8:
					$$=8;
					break;
				case 12:
				case 16:
					$$=16;
					break;
				default:
					char *msg=alloc_msgmem();

					sprintf(msg,"WAVPARS: Illegal size %i bit\n",$1);
					error(msg,ps->wavlineno);
					dealloc_msgmem(msg);
			}
	  }
	;

startat:
	  duration
   ;

duration:
	  expression duration_multiplier
	  {
	 		$$=$1*$2;
	  }
   ;

duration_multiplier:
	  SEC
	  {
	 		$$=1.0;
	  }
	| MSEC
	  {
	 		$$=0.001;
	  }
	;

channelid:
	  MONO
	  {
	 		if (WAVFile.GetMode()!=MODE_MONO)
			{
				warning("WAVPARS: STEREO mode selected! Using LEFT channel!\n",ps->wavlineno);
				$$=CHANNEL_LEFT;
			}
			else
				$$=CHANNEL_MONO;
	  }
	| LEFTCH
	  {
	 		if (WAVFile.GetMode()!=MODE_STEREO)
			{
				warning("WAVPARS: MONO mode selected! Ignoring channel selection!\n",ps->wavlineno);
				$$=CHANNEL_MONO;
			}
			else
				$$=CHANNEL_LEFT;
	  }
	| RIGHTCH
	  {
	 		if (WAVFile.GetMode()!=MODE_STEREO)
			{
				warning("WAVPARS: MONO mode selected! Ignoring channel selection!\n",ps->wavlineno);
				$$=CHANNEL_MONO;
			}
			else
				$$=CHANNEL_RIGHT;
	  }
	;

delimiter:
	  ';'
	| /* empty */
	  {
	 		error("WAVPARS: missing semicolon !",ps->wavlineno);
	  }
	;

on_off:
	  ON
	  {
			$$=SWITCH_ON;
	  }
	| OFF
	  {
			$$=SWITCH_OFF;
	  }
	;

%%

void generate_tone_proc(char channelid,VALUETYPE frequency,VALUETYPE amplitude,VALUETYPE startat, VALUETYPE duration)
{
	float compression_ratio=DEF_COMPRESSION;

	if (verbose==SWITCH_ON)
	{
		switch(channelid)
		{
			case CHANNEL_MONO:
				printf("Generating tone mono :\n");
				break;
			case CHANNEL_LEFT:
				printf("Generating tone on left channel :\n");
				break;
			case CHANNEL_RIGHT:
				printf("Generating tone on right channel:\n");
				break;
		}
		printf("   frequency         : %7.3f Hz\n",frequency);
		printf("   amplitude         : %7.3f\n",amplitude);
		printf("   start at          : %7.3f sec\n",startat);
		printf("   duration          : %7.3f sec\n",duration);
		printf("\n");
		printf("   interval          :[%7.3f - %7.3f] sec\n",startat,startat+duration);

		if (amplitude > 0.0)
		{
			if (autoscale==SWITCH_ON)
			{
				printf("Calculating amplitude compression ratio...");
				compression_ratio=WAVFile.CalcToneCompression(channelid,frequency,amplitude,startat,duration,compression_ratio);
				if (verbose==SWITCH_ON)
					printf(" --> compression ratio = %7.2f\n",compression_ratio);
			}
			printf("Generating tone...\n");
			WAVFile.GenTone(channelid,frequency,amplitude,startat,duration,compression_ratio);
		}
	}
}

void WAV_declare_longvariable(char *varid,VALUETYPE initvalue,unsigned int lineno)
{
	SimpleVariable *help;

	if (WAV_LongVariables.Search(varid,VAR_SIMPLE)!=NULL)
	{
		char *msg=alloc_msgmem();

		sprintf(msg,"WAVPARS: variable %s already declared !",varid);
		warning(msg,lineno);
		dealloc_msgmem(msg);
	}
	else
	{
		if ((help=new SimpleVariable(varid,initvalue))==NULL)
			fatal("WAVPARS: out of memory!!",lineno);
		WAV_LongVariables.Insert(help);
	}
}
